/**
 * @file main.c
 * @brief Main routine
 *
 * @section License
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Copyright (C) 2010-2021 Oryx Embedded SARL. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * @author Oryx Embedded SARL (www.oryx-embedded.com)
 * @version 2.1.2
 **/

//Dependencies
#include <stdlib.h>
#include "stm32f4xx.h"
#include "stm32f4xx_hal.h"
#include "Board_LED.h"
#include "Board_Buttons.h"
#include "Board_GLCD.h"
#include "GLCD_Config.h"

//CMSIS-RTOS and network stack dependencies
#include "cmsis_os.h"
#include "rl_net.h"

//TLS library dependencies
#include "tls.h"
#include "tls_cipher_suites.h"
#include "tls_ticket.h"
#include "tls_misc.h"
#include "rng/yarrow.h"
#include "resource_manager.h"
#include "debug.h"

//Application configuration
#define APP_SERVER_PORT 443
#define APP_SERVER_MAX_CONNECTIONS 4
#define APP_SERVER_TIMEOUT 10000
#define APP_SERVER_CERT "certs/server_cert.pem"
#define APP_SERVER_KEY "certs/server_key.pem"
#define APP_CA_CERT "certs/ca_cert.pem"

//Global variables
uint_t lcdLine = 0;
uint_t lcdColumn = 0;
uint_t hitCounter = 0;

RNG_HandleTypeDef RNG_Handle;

TlsCache *tlsCache;
TlsTicketContext tlsTicketContext;
YarrowContext yarrowContext;
uint8_t seed[32];

//Thread definitions
void tlsServerThread(void const *arg);
osThreadDef(tlsServerThread, osPriorityNormal, 1, 1000);

void userThread(void const *arg);
osThreadDef(userThread, osPriorityNormal, 1, 1000);

void blinkThread(void const *arg);
osThreadDef(blinkThread, osPriorityNormal, 1, 200);

//RTOS object definitions
osSemaphoreId connectionSemaphore;
osSemaphoreDef(connectionSemaphore);

//External variables
extern GLCD_FONT GLCD_Font_16x24;

//Forward declaration of functions
void tlsClientThread(void const *arg);

size_t dumpArray(char_t *buffer, const uint8_t *data, size_t length);

error_t tlsSendCallback(TlsIoHandle socket, const void *data,
   size_t length, size_t *written, uint_t flags);

error_t tlsReceiveCallback(TlsIoHandle socket, void *data,
   size_t size, size_t *received, uint_t flags);


/**
 * @brief Set cursor location
 * @param[in] line Line number
 * @param[in] column Column number
 **/

void lcdSetCursor(uint_t line, uint_t column)
{
   lcdLine = MIN(line, 10);
   lcdColumn = MIN(column, 20);
}


/**
 * @brief Write a character to the LCD display
 * @param[in] c Character to be written
 **/

void lcdPutChar(char_t c)
{
   if(c == '\r')
   {
      lcdColumn = 0;
   }
   else if(c == '\n')
   {
      lcdColumn = 0;
      lcdLine++;
   }
   else if(lcdLine < 10 && lcdColumn < 20)
   {
      //Display current character
      GLCD_DrawChar(lcdColumn * 16, lcdLine * 24, c);

      //Advance the cursor position
      if(++lcdColumn >= 20)
      {
         lcdColumn = 0;
         lcdLine++;
      }
   }
}


/**
 * @brief System clock configuration
 **/

void SystemClock_Config(void)
{
   RCC_OscInitTypeDef RCC_OscInitStruct;
   RCC_ClkInitTypeDef RCC_ClkInitStruct;

   //Enable Power Control clock
   __HAL_RCC_PWR_CLK_ENABLE();

   //The voltage scaling allows optimizing the power consumption when the
   //device is clocked below the maximum system frequency
   __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE1);

   //Enable HSE Oscillator and activate PLL with HSE as source
   RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
   RCC_OscInitStruct.HSEState = RCC_HSE_ON;
   RCC_OscInitStruct.HSIState = RCC_HSI_OFF;
   RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
   RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
   RCC_OscInitStruct.PLL.PLLM = 25;
   RCC_OscInitStruct.PLL.PLLN = 336;
   RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV2;
   RCC_OscInitStruct.PLL.PLLQ = 7;
   HAL_RCC_OscConfig(&RCC_OscInitStruct);

   //Select PLL as system clock source and configure the HCLK, PCLK1 and PCLK2
   //clocks dividers
   RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_SYSCLK | RCC_CLOCKTYPE_HCLK |
      RCC_CLOCKTYPE_PCLK1 | RCC_CLOCKTYPE_PCLK2;
   RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
   RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
   RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV4;
   RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV2;
   HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_5);

   //Revision Z devices?
   if (HAL_GetREVID() == 0x1001)
   {
      //Enable Flash prefetch
      __HAL_FLASH_PREFETCH_BUFFER_ENABLE();
   }
}


/**
 * @brief User thread
 * @param[in] param Unused parameter
 **/

void userThread(void const *arg)
{
   uint8_t ipv4Addr[NET_ADDR_IP4_LEN];
   char_t buffer[20];

   //Initialize LCD display
   lcdSetCursor(2, 0);
   printf("IPv4 Addr\r\n");
   lcdSetCursor(5, 0);
   printf("Subnet Mask\r\n");

   //Endless loop
   while(1)
   {
      //Move cursor
      lcdSetCursor(3, 0);

      //Retrieve IPv4 address
      netIF_GetOption(NET_IF_CLASS_ETH, netIF_OptionIP4_Address,
         ipv4Addr, NET_ADDR_IP4_LEN);

      //Display IPv4 address
      netIP_ntoa(NET_ADDR_IP4, ipv4Addr, buffer, sizeof(buffer));
      printf("%s        \r\n", buffer);

      //Move cursor
      lcdSetCursor(6, 0);

      //Retrieve subnet mask
      netIF_GetOption(NET_IF_CLASS_ETH, netIF_OptionIP4_SubnetMask,
         ipv4Addr, NET_ADDR_IP4_LEN);

      //Display subnet mask
      netIP_ntoa(NET_ADDR_IP4, ipv4Addr, buffer, sizeof(buffer));
      printf("%s        \r\n", buffer);

      //Loop delay
      osDelay(1000);
   }
}


/**
 * @brief LED blinking thread
 * @param[in] arg Unused parameter
 **/

void blinkThread(void const *arg)
{
   //Endless loop
   while(1)
   {
      LED_On(0);
      osDelay(100);
      LED_Off(0);
      osDelay(900);
   }
}


/**
 * @brief Main entry point
 * @return Unused value
 **/

int main(void)
{
   error_t error;
   int_t i;
   uint32_t value;
   osThreadId id;
   netStatus status;

   //HAL library initialization
   HAL_Init();
   //Configure the system clock
   SystemClock_Config();

   //Initialize kernel
   osKernelInitialize();
   //Configure debug UART
   debugInit(115200);

   //Start-up message
   TRACE_INFO("\r\n");
   TRACE_INFO("**********************************\r\n");
   TRACE_INFO("*** CycloneSSL TLS Server Demo ***\r\n");
   TRACE_INFO("**********************************\r\n");
   TRACE_INFO("Copyright: 2010-2021 Oryx Embedded SARL\r\n");
   TRACE_INFO("Compiled: %s %s\r\n", __DATE__, __TIME__);
   TRACE_INFO("Target: STM32F407\r\n");
   TRACE_INFO("\r\n");

   //LED configuration
   LED_Initialize();
   //Clear LEDs
   LED_Off(0);
   LED_Off(1);
   LED_Off(2);
   LED_Off(3);
   LED_Off(4);
   LED_Off(5);
   LED_Off(6);
   LED_Off(7);

   //Initialize buttons
   Buttons_Initialize();

   //Initialize graphic LCD display
   GLCD_Initialize();
   GLCD_SetForegroundColor(GLCD_COLOR_WHITE);
   GLCD_SetBackgroundColor(GLCD_COLOR_BLUE);
   GLCD_ClearScreen();
   GLCD_SetFont(&GLCD_Font_16x24);

   //Welcome message
   lcdSetCursor(0, 0);
   printf("TLS Server Demo\r\n");

   //Enable RNG peripheral clock
   __HAL_RCC_RNG_CLK_ENABLE();
   //Initialize RNG
   RNG_Handle.Instance = RNG;
   HAL_RNG_Init(&RNG_Handle);

   //Generate a random seed
   for(i = 0; i < 32; i += 4)
   {
      //Get 32-bit random value
      HAL_RNG_GenerateRandomNumber(&RNG_Handle, &value);

      //Copy random value
      seed[i] = value & 0xFF;
      seed[i + 1] = (value >> 8) & 0xFF;
      seed[i + 2] = (value >> 16) & 0xFF;
      seed[i + 3] = (value >> 24) & 0xFF;
   }

   //PRNG initialization
   error = yarrowInit(&yarrowContext);
   //Any error to report?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize PRNG!\r\n");
   }

   //Properly seed the PRNG
   error = yarrowSeed(&yarrowContext, seed, sizeof(seed));
   //Any error to report?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to seed PRNG!\r\n");
   }

   //TCP/IP stack initialization
   status = netInitialize();
   //Any error to report?
   if(status != netOK)
   {
      //Debug message
      TRACE_ERROR("Failed to initialize TCP/IP stack!\r\n");
   }

   //Create a thread to handle incoming requests
   id = osThreadCreate(osThread(tlsServerThread), NULL);
   //Failed to create the task?
   if(id == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create thread!\r\n");
   }

   //Create the user thread
   id = osThreadCreate(osThread(userThread), NULL);
   //Failed to create the thread?
   if(id == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create thread!\r\n");
   }

   //Create a thread to blink the LED
   id = osThreadCreate(osThread(blinkThread), NULL);
   //Failed to create the thread?
   if(id == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create thread!\r\n");
   }

   //Start the kernel
   osKernelStart();

   //This function should never return
   return 0;
}


/**
 * @brief TLS server task
 * @param param[in] arg Not used
 **/

void tlsServerThread(const void *arg)
{
   uint_t counter;
   int ret;
   int serverSocket;
   int clientSocket;
   int clientAddrLen;
   SOCKADDR_IN serverAddr;
   SOCKADDR_IN clientAddr;
   osThreadDef_t threadDef;
   osThreadId id;

   //Create a semaphore to limit the number of simultaneous connections
   connectionSemaphore = osSemaphoreCreate(osSemaphore(connectionSemaphore), 4);
   //Any error to report?
   if(connectionSemaphore == NULL)
   {
      //Debug message
      TRACE_ERROR("Failed to create semaphore!\r\n");
   }

   //TLS session cache initialization
   tlsCache = tlsInitCache(16);
   //Failed to initialize TLS cache?
   if(tlsCache == NULL)
   {
      //Debug message
      TRACE_ERROR("TLS cache initialization failed!\r\n");
   }

#if (TLS_TICKET_SUPPORT == ENABLED)
   //Initialize ticket encryption context
   error = tlsInitTicketContext(&tlsTicketContext);
   //Any error to report?
   if(error)
   {
      //Debug message
      TRACE_ERROR("Failed to bind socket!\r\n");
   }
#endif

   //Open a socket
   serverSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
   //Failed to open socket?
   if(serverSocket < 0)
   {
      //Debug message
      TRACE_ERROR("Cannot open socket!\r\n");
   }

   //Server port (443)
   serverAddr.sin_port = htons(APP_SERVER_PORT);
   serverAddr.sin_family = PF_INET;
   serverAddr.sin_addr.s_addr = INADDR_ANY;

   //Bind the socket to the relevant port number
   ret = bind(serverSocket, (SOCKADDR *) &serverAddr, sizeof(serverAddr));
   //Any error to report?
   if(ret < 0)
   {
      //Debug message
      TRACE_ERROR("Failed to bind socket!\r\n");
   }

   //Place socket in listening state
   ret = listen(serverSocket, 1);
   //Any error to report?
   if(ret < 0)
   {
      //Debug message
      TRACE_ERROR("Failed to enter listening state!\r\n");
   }

   //Process incoming connections to the server
   for(counter = 1; ; counter++)
   {
      //Debug message
      TRACE_INFO("\r\n\r\n");
      TRACE_INFO("Waiting for an incoming connection...\r\n\r\n");

      //Limit the number of simultaneous connections to the HTTP server
      osSemaphoreWait(connectionSemaphore, osWaitForever);

      //Accept an incoming connection
      clientSocket = accept(serverSocket, (SOCKADDR *) &clientAddr, &clientAddrLen);

      //Make sure the socket handle is valid
      if(clientSocket >= 0)
      {
         //Debug message
         TRACE_INFO("Connection #%u established with client %u.%u.%u.%u port %u...\r\n",
            hitCounter, clientAddr.sin_addr.s_b1, clientAddr.sin_addr.s_b2,
            clientAddr.sin_addr.s_b3, clientAddr.sin_addr.s_b4,
            ntohs(clientAddr.sin_port));

         //thread parameters
         threadDef.pthread = tlsClientThread;
         threadDef.tpriority = osPriorityNormal;
         threadDef.instances = 1;
         threadDef.stacksize = 2000;

         //Create a thread to service the client connection
         id = osThreadCreate(&threadDef, (void *) clientSocket);

         //Did we encounter an error?
         if(id == NULL)
         {
            //Debug message
            TRACE_ERROR("Failed to create task!\r\n");

            //Close socket
            closesocket(clientSocket);
            //Release semaphore
            osSemaphoreRelease(connectionSemaphore);
         }
      }
   }
}


/**
 * @brief TLS client task
 * @param param[in] arg Client socket
 **/

void tlsClientThread(const void *arg)
{
   error_t error;
   size_t n;
   const char_t *serverCert;
   size_t serverCertLen;
   const char_t *serverKey;
   size_t serverKeyLen;
   const char_t *trustedCaList;
   size_t trustedCaListLen;
   int clientSocket;
   TlsContext *tlsContext;
   char_t buffer[512];

   //Variable initialization
   tlsContext = NULL;

   //Retrieve socket handle
   clientSocket = (int) arg;

   //Start of exception handling block
   do
   {
      //TLS context initialization
      tlsContext = tlsInit();
      //Failed to initialize TLS context?
      if(tlsContext == NULL)
      {
         //Report an error
         error = ERROR_OUT_OF_MEMORY;
         //Exit immediately
         break;
      }

      //Select server operation mode
      error = tlsSetConnectionEnd(tlsContext, TLS_CONNECTION_END_SERVER);
      //Any error to report?
      if(error)
         break;

      //Set send and receive callbacks (I/O abstraction layer)
      error = tlsSetIoCallbacks(tlsContext, (TlsIoHandle) clientSocket,
         tlsSendCallback, tlsReceiveCallback);
      //Any error to report?
      if(error)
         break;

      //Set TX and RX buffer size
      error = tlsSetBufferSize(tlsContext, 4096, 4096);
      //Any error to report?
      if(error)
         break;

      //Set the PRNG algorithm to be used
      error = tlsSetPrng(tlsContext, YARROW_PRNG_ALGO, &yarrowContext);
      //Any error to report?
      if(error)
         break;

      //Session cache that will be used to save/resume TLS sessions
      error = tlsSetCache(tlsContext, tlsCache);
      //Any error to report?
      if(error)
         break;

#if (TLS_TICKET_SUPPORT == ENABLED)
      //Enable session ticket mechanism
      error = tlsSetTicketCallbacks(tlsContext, tlsEncryptTicket,
         tlsDecryptTicket, &tlsTicketContext);
      //Any error to report?
      if(error)
         break;
#endif

      //Client authentication is not required
      error = tlsSetClientAuthMode(tlsContext, TLS_CLIENT_AUTH_NONE);
      //Any error to report?
      if(error)
         break;

      //Get server's certificate
      error = resGetData(APP_SERVER_CERT, (const uint8_t **) &serverCert,
         &serverCertLen);
      //Any error to report?
      if(error)
         break;

      //Get server's private key
      error = resGetData(APP_SERVER_KEY, (const uint8_t **) &serverKey,
         &serverKeyLen);
      //Any error to report?
      if(error)
         break;

      //Import the server's certificate
      error = tlsAddCertificate(tlsContext, serverCert, serverCertLen,
         serverKey, serverKeyLen);
      //Any error to report?
      if(error)
         break;

      //Get the list of trusted CA certificates
      error = resGetData(APP_CA_CERT, (const uint8_t **) &trustedCaList,
         &trustedCaListLen);
      //Any error to report?
      if(error)
         break;

      //Import the list of trusted CA certificates
      error = tlsSetTrustedCaList(tlsContext, trustedCaList, trustedCaListLen);
      //Any error to report?
      if(error)
         break;

      //Establish a secure session
      error = tlsConnect(tlsContext);
      //TLS handshake failure?
      if(error)
         break;

      //Debug message
      TRACE_INFO("\r\n");
      TRACE_INFO("HTTP request:\r\n");

      //Read HTTP request
      while(1)
      {
         //Read a complete line
         error = tlsRead(tlsContext, buffer, sizeof(buffer) - 1, &n,
            TLS_FLAG_BREAK_CRLF);
         //Any error to report?
         if(error)
            break;

         //Properly terminate the string with a NULL character
         buffer[n] = '\0';
         //Dump HTTP request
         TRACE_INFO("%s", buffer);

         //The end of the header has been reached?
         if(!strcmp(buffer, "\r\n"))
            break;
      }

      //Propagate exception if necessary
      if(error)
         break;

      //Debug message
      TRACE_INFO("HTTP response:\r\n");

      //Format response
      n = 0;
      n += sprintf(buffer + n, "HTTP/1.0 200 OK\r\n");
      n += sprintf(buffer + n, "Content-Type: text/html\r\n");
      n += sprintf(buffer + n, "\r\n");

      n += sprintf(buffer + n, "<!DOCTYPE html>\r\n");
      n += sprintf(buffer + n, "<html>\r\n");
      n += sprintf(buffer + n, "<head>\r\n");
      n += sprintf(buffer + n, "  <title>Oryx Embedded - CycloneSSL TLS Server Demo</title>\r\n");
      n += sprintf(buffer + n, "  <style>\r\n");
      n += sprintf(buffer + n, "    body {font-family: monospace; font-size: 13px;}\r\n");
      n += sprintf(buffer + n, "    table {border-width: 1px; border-style: ouset; border-collapse: collapse;}\r\n");
      n += sprintf(buffer + n, "    td {border-width: 1px; border-style: inset; padding: 3px;}\r\n");
      n += sprintf(buffer + n, "  </style>\r\n");
      n += sprintf(buffer + n, "</head>\r\n");
      n += sprintf(buffer + n, "<body>\r\n");
      n += sprintf(buffer + n, "  <p>Welcome to the CycloneSSL TLS server!</p>\r\n");
      n += sprintf(buffer + n, "  <table>\r\n");

      //Debug message
      TRACE_INFO("%s\r\n", buffer);

      //Send response to the client
      error = tlsWrite(tlsContext, buffer, n, NULL, 0);
      //Any error to report?
      if(error)
         break;

      //Format response
      n = 0;
      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Hit counter</td>\r\n");
      n += sprintf(buffer + n, "    <td>%d</td>\r\n", ++hitCounter);
      n += sprintf(buffer + n, "  </tr>\r\n");

      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Server version</td>\r\n");
      n += sprintf(buffer + n, "    <td>%s</td>\r\n",
         tlsGetVersionName(tlsContext->versionMax));
      n += sprintf(buffer + n, "  </tr>\r\n");

      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Client version</td>\r\n");
      n += sprintf(buffer + n, "    <td>%s</td>\r\n",
         tlsGetVersionName(MAX(tlsContext->clientVersion, tlsContext->version)));
      n += sprintf(buffer + n, "  </tr>\r\n");

      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Negotiated version</td>\r\n");
      n += sprintf(buffer + n, "    <td>%s</td>\r\n",
         tlsGetVersionName(tlsContext->version));
      n += sprintf(buffer + n, "  </tr>\r\n");

      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Cipher suite</td>\r\n");
      n += sprintf(buffer + n, "    <td>%s</td>\r\n",
         tlsGetCipherSuiteName(tlsContext->cipherSuite.identifier));
      n += sprintf(buffer + n, "  </tr>\r\n");

      //Debug message
      TRACE_INFO("%s\r\n", buffer);

      //Send response to the client
      error = tlsWrite(tlsContext, buffer, n, NULL, 0);
      //Any error to report?
      if(error)
         break;

      //Format response
      n = 0;
      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Client random</td>\r\n");
      n += sprintf(buffer + n, "      <td>\r\n");
      n += sprintf(buffer + n, "        ");
      n += dumpArray(buffer + n, (uint8_t *) &tlsContext->clientRandom, 32);
      n += sprintf(buffer + n, "\r\n");
      n += sprintf(buffer + n, "      </td>\r\n");
      n += sprintf(buffer + n, "  </tr>\r\n");

      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Server random</td>\r\n");
      n += sprintf(buffer + n, "      <td>\r\n");
      n += sprintf(buffer + n, "        ");
      n += dumpArray(buffer + n, (uint8_t *) &tlsContext->serverRandom, 32);
      n += sprintf(buffer + n, "\r\n");
      n += sprintf(buffer + n, "      </td>\r\n");
      n += sprintf(buffer + n, "  </tr>\r\n");

      //Debug message
      TRACE_INFO("%s\r\n", buffer);

      //Send response to the client
      error = tlsWrite(tlsContext, buffer, n, NULL, 0);
      //Any error to report?
      if(error)
         break;

      //Format response
      n = 0;
      n += sprintf(buffer + n, "  <tr>\r\n");
      n += sprintf(buffer + n, "    <td>Session ID</td>\r\n");
      n += sprintf(buffer + n, "      <td>\r\n");
      n += sprintf(buffer + n, "        ");
      n += dumpArray(buffer + n, tlsContext->sessionId, tlsContext->sessionIdLen);
      n += sprintf(buffer + n, "\r\n");
      n += sprintf(buffer + n, "      </td>\r\n");
      n += sprintf(buffer + n, "  </tr>\r\n");

      n += sprintf(buffer + n, "  </table>\r\n");
      n += sprintf(buffer + n, "</body>\r\n");
      n += sprintf(buffer + n, "</html>\r\n");

      //Debug message
      TRACE_INFO("%s\r\n", buffer);

      //Send response to the client
      error = tlsWrite(tlsContext, buffer, n, NULL, 0);
      //Any error to report?
      if(error)
         break;

      //Terminate TLS session
      error = tlsShutdown(tlsContext);
      //Any error to report?
      if(error)
         break;

      //End of exception handling block
   } while(0);

   //Release TLS context
   if(tlsContext != NULL)
   {
      tlsFree(tlsContext);
   }

   //Close socket
   if(clientSocket >= 0)
   {
      closesocket(clientSocket);
   }

   //Debug message
   TRACE_INFO("Connection closed...\r\n");

   //Release semaphore
   osSemaphoreRelease(connectionSemaphore);

   //Kill ourselves
   osThreadTerminate(NULL);
}


/**
 * @brief Display the contents of an array
 * @param[out] buffer Output buffer where to format the resulting string
 * @param[in] data Pointer to the data array
 * @param[in] length Number of bytes in the array
 * @return Length of the resulting string
 **/

size_t dumpArray(char_t *buffer, const uint8_t *data, size_t length)
{
   size_t i;
   size_t n;

   //Variable initialization
   n = 0;

   //Properly terminate the string
   buffer[0] = '\0';

   //Process input data
   for(i = 0; i < length; i++)
   {
      //Beginning of a new line?
      if(i != 0 && (i % 16) == 0)
      {
         n += sprintf(buffer + n, "\r\n        ");
      }

      //Display current data byte
      n += sprintf(buffer + n, "%02" PRIX8 " ", data[i]);
   }

   //Return the length of the resulting string
   return n;
}


/**
 * @brief TLS send callback (I/O abstraction layer)
 * @param[in] socket Handle that identifies a socket
 * @param[in] data Pointer to a buffer containing the data to be transmitted
 * @param[in] length Number of data bytes to send
 * @param[out] written Number of bytes that have been transmitted
 * @param[in] flags Unused parameter
 * @return Error code
 **/

error_t tlsSendCallback(TlsIoHandle socket, const void *data,
   size_t length, size_t *written, uint_t flags)
{
   int_t n;
   error_t error;

   //Send data
   n = send((int) socket, data, length, 0);

   //Check return value
   if(n > 0)
   {
      //Total number of data that have been written
      *written = n;
      //Successful write operation
      error = NO_ERROR;
   }
   else
   {
      //Timeout error?
      if(n == BSD_ERROR_WOULDBLOCK || n == BSD_ERROR_TIMEOUT)
         error = ERROR_TIMEOUT;
      else
         error = ERROR_WRITE_FAILED;
   }

   //Return status code
   return error;
}


/**
 * @brief TLS receive callback (I/O abstraction layer)
 * @param[in] socket Handle that identifies a socket
 * @param[out] data Buffer where to store the incoming data
 * @param[in] size Maximum number of bytes that can be received
 * @param[out] received Number of bytes that have been received
 * @param[in] flags Unused parameter
 * @return Error code
 **/

error_t tlsReceiveCallback(TlsIoHandle socket, void *data,
   size_t size, size_t *received, uint_t flags)
{
   int_t n;
   error_t error;

   //Send data
   n = recv((int) socket, data, size, 0);

   //Check return value
   if(n > 0)
   {
      //Total number of data that have been received
      *received = n;
      //Successful write operation
      error = NO_ERROR;
   }
   else
   {
      //Timeout error?
      if(n == BSD_ERROR_WOULDBLOCK || n == BSD_ERROR_TIMEOUT)
         error = ERROR_TIMEOUT;
      else
         error = ERROR_READ_FAILED;
   }

   //Return status code
   return error;
}
